/* APNG Disassembler 2.7
 *
 * Deconstructs APNG files into individual frames.
 *
 * http://apngdis.sourceforge.net
 *
 * Copyright (c) 2010-2014 Max Stepin
 * maxst at users.sourceforge.net
 *
 * zlib license
 * ------------
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include "png.h"     /* original (unpatched) libpng is ok */
#include "zlib.h"
using namespace std;

#define notabc(c) ((c) < 65 || (c) > 122 || ((c) > 90 && (c) < 97))

#define id_IHDR 0x52444849
#define id_acTL 0x4C546361
#define id_fcTL 0x4C546366
#define id_IDAT 0x54414449
#define id_fdAT 0x54416466
#define id_IEND 0x444E4549

struct CHUNK { unsigned char * p; unsigned int size; };

struct APNGFrame
{
  unsigned char * p;
  unsigned int w, h;
  unsigned int delay_num, delay_den;
  png_bytep * rows;
};

png_structp   g_png;
png_infop     g_info;
unsigned char g_header[8] = {137, 80, 78, 71, 13, 10, 26, 10};
unsigned char g_footer[12] = {0, 0, 0, 0, 73, 69, 78, 68, 174, 66, 96, 130};
CHUNK         g_chunkIHDR;
vector<CHUNK> g_chunksInfo;
vector<APNGFrame> frames;

void info_fn(png_structp png_ptr, png_infop info_ptr)
{
  png_set_expand(png_ptr);
  png_set_strip_16(png_ptr);
  png_set_gray_to_rgb(png_ptr);
  png_set_add_alpha(png_ptr, 0xff, PNG_FILLER_AFTER);
  (void)png_set_interlace_handling(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
}

void row_fn(png_structp png_ptr, png_bytep new_row, png_uint_32 row_num, int pass)
{
  APNGFrame * frame = (APNGFrame *)png_get_progressive_ptr(png_ptr);
  png_progressive_combine_row(png_ptr, frame->rows[row_num], new_row);
}

void compose_frame(unsigned char ** rows_dst, unsigned char ** rows_src, unsigned char bop, unsigned int x, unsigned int y, unsigned int w, unsigned int h)
{
  unsigned int  i, j;
  int u, v, al;

  for (j=0; j<h; j++)
  {
    unsigned char * sp = rows_src[j];
    unsigned char * dp = rows_dst[j+y] + x*4;

    if (bop == 0)
      memcpy(dp, sp, w*4);
    else
    for (i=0; i<w; i++, sp+=4, dp+=4)
    {
      if (sp[3] == 255)
        memcpy(dp, sp, 4);
      else
      if (sp[3] != 0)
      {
        if (dp[3] != 0)
        {
          u = sp[3]*255;
          v = (255-sp[3])*dp[3];
          al = u + v;
          dp[0] = (sp[0]*u + dp[0]*v)/al;
          dp[1] = (sp[1]*u + dp[1]*v)/al;
          dp[2] = (sp[2]*u + dp[2]*v)/al;
          dp[3] = al/255;
        }
        else
          memcpy(dp, sp, 4);
      }
    }
  }
}

unsigned int read_chunk(FILE * f, CHUNK * pChunk)
{
  unsigned char len[4];
  if (fread(&len, 4, 1, f) == 1)
  {
    pChunk->size = png_get_uint_32(len) + 12;
    pChunk->p = new unsigned char[pChunk->size];
    memcpy(pChunk->p, len, 4);
    if (fread(pChunk->p + 4, pChunk->size - 4, 1, f) == 1)
      return *(unsigned int *)(pChunk->p + 4);
  }
  return 0;
}

void processing_start(void * frame_ptr, bool hasInfo)
{
  g_png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  g_info = png_create_info_struct(g_png);
  if (!g_png || !g_info)
    return;

  if (setjmp(png_jmpbuf(g_png)))
  {
    png_destroy_read_struct(&g_png, &g_info, 0);
    return;
  }

  png_set_crc_action(g_png, PNG_CRC_QUIET_USE, PNG_CRC_QUIET_USE);
  png_set_progressive_read_fn(g_png, frame_ptr, info_fn, row_fn, NULL);
          
  png_process_data(g_png, g_info, g_header, 8);
  png_process_data(g_png, g_info, g_chunkIHDR.p, g_chunkIHDR.size);

  if (hasInfo)
    for (unsigned int i=0; i<g_chunksInfo.size(); ++i)
      png_process_data(g_png, g_info, g_chunksInfo[i].p, g_chunksInfo[i].size);
}

void processing_data(unsigned char * p, unsigned int size)
{
  if (!g_png || !g_info)
    return;

  if (setjmp(png_jmpbuf(g_png)))
  {
    png_destroy_read_struct(&g_png, &g_info, 0);
    return;
  }

  png_process_data(g_png, g_info, p, size);
}

int processing_finish()
{
  if (!g_png || !g_info)
    return 1;

  if (setjmp(png_jmpbuf(g_png)))
  {
    png_destroy_read_struct(&g_png, &g_info, 0);
    return 1;
  }

  png_process_data(g_png, g_info, g_footer, 12);
  png_destroy_read_struct(&g_png, &g_info, 0);

  return 0;
}

int load_apng(char * szIn)
{
  FILE         *  f;
  unsigned int    id, i, j, w, h, w0, h0, x0, y0;
  unsigned int    delay_num, delay_den, dop, bop, rowbytes, imagesize;
  CHUNK           chunk;
  unsigned char   sig[8];
  bool            isAnimated = false;
  bool            skipFirst = false;
  bool            hasInfo = false;
  APNGFrame       frameRaw = {0};
  APNGFrame       frameCur = {0};
  APNGFrame       frameNext = {0};
  int             res = -1;

  printf("Reading '%s'...\n", szIn);

  if ((f = fopen(szIn, "rb")) != 0)
  {
    if (fread(sig, 1, 8, f) == 8 && memcmp(sig, g_header, 8) == 0)
    {
      id = read_chunk(f, &g_chunkIHDR);

      if (id == id_IHDR && g_chunkIHDR.size == 25)
      {
        w0 = w = png_get_uint_32(g_chunkIHDR.p + 8);
        h0 = h = png_get_uint_32(g_chunkIHDR.p + 12);
        x0 = 0;
        y0 = 0;
        delay_num = 1;
        delay_den = 10;
        dop = 0;
        bop = 0;
        rowbytes = w * 4;
        imagesize = h * rowbytes;

        frameRaw.p = new unsigned char[imagesize];
        frameRaw.rows = new png_bytep[h * sizeof(png_bytep)];
        for (j=0; j<h; ++j)
          frameRaw.rows[j] = frameRaw.p + j * rowbytes;

        frameCur.w = w;
        frameCur.h = h;
        frameCur.p = new unsigned char[imagesize];
        frameCur.rows = new png_bytep[h * sizeof(png_bytep)];
        for (j=0; j<h; ++j)
          frameCur.rows[j] = frameCur.p + j * rowbytes;

        processing_start((void *)&frameRaw, hasInfo);

        while ( !feof(f) )
        {
          id = read_chunk(f, &chunk);

          if (id == id_acTL && !hasInfo && !isAnimated)
          {
            isAnimated = true;
            skipFirst = true;
          }
          else
          if (id == id_fcTL && (!hasInfo || isAnimated))
          {
            if (hasInfo)
            {
              if (!processing_finish())
              {
                frameNext.p = new unsigned char[imagesize];
                frameNext.rows = new png_bytep[h * sizeof(png_bytep)];
                for (j=0; j<h; ++j)
                  frameNext.rows[j] = frameNext.p + j * rowbytes;

                if (dop == 2)
                  memcpy(frameNext.p, frameCur.p, imagesize);

                compose_frame(frameCur.rows, frameRaw.rows, bop, x0, y0, w0, h0);
                frameCur.delay_num = delay_num;
                frameCur.delay_den = delay_den;

                frames.push_back(frameCur);

                if (dop != 2)
                {
                  memcpy(frameNext.p, frameCur.p, imagesize);
                  if (dop == 1)
                    for (j=0; j<h0; j++)
                      memset(frameNext.rows[y0 + j] + x0*4, 0, w0*4);
                }
                frameCur.p = frameNext.p;
                frameCur.rows = frameNext.rows;
              }
              else
              {
                delete[] frameCur.rows;
                delete[] frameCur.p;
                delete[] chunk.p;
                break;
              }
            }

            // At this point the old frame is done. Let's start a new one.
            w0 = png_get_uint_32(chunk.p + 12);
            h0 = png_get_uint_32(chunk.p + 16);
            x0 = png_get_uint_32(chunk.p + 20);
            y0 = png_get_uint_32(chunk.p + 24);
            delay_num = png_get_uint_16(chunk.p + 28);
            delay_den = png_get_uint_16(chunk.p + 30);
            dop = chunk.p[32];
            bop = chunk.p[33];

            if (hasInfo)
            {
              memcpy(g_chunkIHDR.p + 8, chunk.p + 12, 8);
              processing_start((void *)&frameRaw, hasInfo);
            }
            else
              skipFirst = false;

            if (frames.size() == (skipFirst ? 1 : 0))
            {
              bop = 0;
              if (dop == 2)
                dop = 1;
            }
          }
          else
          if (id == id_IDAT)
          {
            hasInfo = true;
            processing_data(chunk.p, chunk.size);
          }
          else
          if (id == id_fdAT && isAnimated)
          {
            png_save_uint_32(chunk.p + 4, chunk.size - 16);
            memcpy(chunk.p + 8, "IDAT", 4);
            processing_data(chunk.p + 4, chunk.size - 4);
          }
          else
          if (id == id_IEND)
          {
            if (hasInfo && !processing_finish())
            {
              compose_frame(frameCur.rows, frameRaw.rows, bop, x0, y0, w0, h0);
              frameCur.delay_num = delay_num;
              frameCur.delay_den = delay_den;
              frames.push_back(frameCur);
            }
            else
            {
              delete[] frameCur.rows;
              delete[] frameCur.p;
            }
            delete[] chunk.p;
            break;
          }
          else
          if (notabc(chunk.p[4]) || notabc(chunk.p[5]) || notabc(chunk.p[6]) || notabc(chunk.p[7]))
          {
            delete[] chunk.p;
            break;
          }
          else
          if (!hasInfo)
          {
            processing_data(chunk.p, chunk.size);
            g_chunksInfo.push_back(chunk);
            continue;
          }
          delete[] chunk.p;
        }
        delete[] frameRaw.rows;
        delete[] frameRaw.p;

        if (!frames.empty())
          res = (skipFirst) ? 0 : 1;
      }
    }
    fclose(f);

    for (i=0; i<g_chunksInfo.size(); ++i)
      delete[] g_chunksInfo[i].p;

    g_chunksInfo.clear();
    delete[] g_chunkIHDR.p;
  }

  return res;
}

void save_png(char * szOut, APNGFrame * frame)
{
  FILE * f;
  png_structp  png_ptr  = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  png_infop    info_ptr = png_create_info_struct(png_ptr);

  if (!png_ptr || !info_ptr)
    return;

  if (setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    return;
  }

  if ((f = fopen(szOut, "wb")) != 0)
  {
    png_init_io(png_ptr, f);
    png_set_compression_level(png_ptr, 9);
    png_set_IHDR(png_ptr, info_ptr, frame->w, frame->h, 8, 6, 0, 0, 0);
    png_write_info(png_ptr, info_ptr);
    png_write_image(png_ptr, frame->rows);
    png_write_end(png_ptr, info_ptr);
    fclose(f);
  }
  png_destroy_write_struct(&png_ptr, &info_ptr);
}

void save_txt(char * szOut, APNGFrame * frame)
{
  FILE * f;
  if ((f = fopen(szOut, "wb")) != 0)
  {
    fprintf(f, "delay=%d/%d\n", frame->delay_num, frame->delay_den);
    fclose(f);
  }
}

int main(int argc, char** argv)
{
  unsigned int i, j;
  char * szInput;
  char * szOutPrefix;
  char   szPath[256];
  char   szOut[256];

  printf("\nAPNG Disassembler 2.7\n\n");

  if (argc > 1)
    szInput = argv[1];
  else
  {
    printf("Usage: apngdis anim.png [name]\n");
    return 1;
  }

  strcpy(szPath, szInput);
  for (i=j=0; szPath[i]!=0; i++)
  {
    if (szPath[i] == '\\' || szPath[i] == '/' || szPath[i] == ':')
      j = i+1;
  }
  szPath[j] = 0;

  if (argc > 2)
  {
    szOutPrefix = argv[2];

    for (i=j=0; szOutPrefix[i]!=0; i++)
    {
      if (szOutPrefix[i] == '\\' || szOutPrefix[i] == '/' || szOutPrefix[i] == ':')
        j = i+1;
      if (szOutPrefix[i] == '.')
        szOutPrefix[i] = 0;
    }
    strcat(szPath, szOutPrefix+j);
  }
  else
    strcat(szPath, "apngframe");

  int res = load_apng(szInput);
  if (res < 0)
  {
    printf("load_apng() failed: '%s'\n ", szInput);
    return 1;
  }
  unsigned int num_frames = frames.size();
  unsigned int len = sprintf(szOut, "%d", num_frames);

  for (i=0; i<num_frames; ++i)
  {
    printf("extracting frame %d of %d\n", i+1, num_frames);

    sprintf(szOut, "%s%.*d.png", szPath, len, i+res);
    save_png(szOut, &frames[i]);

    sprintf(szOut, "%s%.*d.txt", szPath, len, i+res);
    save_txt(szOut, &frames[i]);

    delete[] frames[i].rows;
    delete[] frames[i].p;
  }
  frames.clear();

  printf("all done\n");

  return 0;
}
